package arzumify.polyenergy.impl;

import java.util.ArrayList;
import java.util.HashMap;

// It's slightly more optimized now, as in it could realistically run in Minecraft.
// Reset: O(1) constant time, we're just deleting everything
// Add Provider: O(m * k^2) where m is the number of receivers and k is the number of points of presence
// Add Receiver: O(n * k^2) where n is the number of providers and k is the number of points of presence
// Remove Provider: O(n + m) where n is the number of providers and m is the number of matches for the provider
// Remove Receiver: O(m) where m is the number of receivers
// I think we stop the optimisations here, the bottleneck is no longer the matchmaker but Minecraft itself.
// If more optimisation is needed we can try to perform Vector math in a more efficient way, but I think this is good enough.

public class CoordinateMatchMaker {
    private static final ArrayList<ProviderDetails> providers = new ArrayList<>();
    private static final ArrayList<ReceiverDetails> receivers = new ArrayList<>();
    private static final HashMap<ProviderDetails, ArrayList<ReceiverDetails>> matches = new HashMap<>();

    public static void reset() {
        providers.clear();
        receivers.clear();
        matches.clear();
    }

    public static void addProvider(ProviderDetails provider) {
        providers.add(provider);
        matches.put(provider, new ArrayList<>());
        // Search for receivers within range of this provider
        for (ReceiverDetails receiver : receivers) {
            if (provider.isInRange(receiver.pointsOfPresence())) {
                provider.provider().exists(receiver.receiver(), provider.pointsOfPresence());
                matches.get(provider).add(receiver);
            }
        }
    }

    public static void addReceiver(ReceiverDetails receiver) {
        receivers.add(receiver);
        // Search for providers within range of this receiver
        for (ProviderDetails provider : providers) {
            if (provider.isInRange(receiver.pointsOfPresence())) {
                provider.provider().exists(receiver.receiver(), provider.pointsOfPresence());
                matches.get(provider).add(receiver);
            }
        }
    }

    public static void removeProvider(ProviderDetails provider) {
        providers.remove(provider);
        for (ReceiverDetails receiver : matches.get(provider)) {
            receiver.receiver().unready(provider.provider());
        }
        matches.remove(provider);
    }

    public static void removeReceiver(ReceiverDetails receiver) {
        receivers.remove(receiver);
    }
}

