package arzumify.polyenergy.util;

import arzumify.polyenergy.api.EnergyProvider;
import arzumify.polyenergy.api.EnergyReceiver;
import arzumify.polyenergy.impl.CoordinateMatchMaker;
import arzumify.polyenergy.impl.ProviderDetails;
import arzumify.polyenergy.impl.ReceiverDetails;
import net.fabricmc.fabric.api.event.lifecycle.v1.ServerBlockEntityEvents;
import net.minecraft.block.BlockState;
import net.minecraft.block.entity.BlockEntity;
import net.minecraft.block.entity.BlockEntityType;
import net.minecraft.server.world.ServerWorld;
import net.minecraft.util.math.BlockPos;
import net.minecraft.util.math.Vec3i;
import net.minecraft.world.World;

import java.util.ArrayList;
import java.util.concurrent.CopyOnWriteArrayList;

/**
 * A simple battery that can provide and receive energy. Has a range of 1 block in all directions (directly adjacent blocks).
 */
public class SimpleBattery extends BlockEntity implements EnergyProvider, EnergyReceiver, ServerBlockEntityEvents.Unload, ServerBlockEntityEvents.Load {
    private final long capacity;
    private final long inputRate;
    private final long outputRate;
    private final ProviderDetails providerDetails = ProviderDetails.NewSimple(pos, this);
    private final ReceiverDetails receiverDetails = ReceiverDetails.NewSimple(pos, this);
    private final CopyOnWriteArrayList<EnergyProvider> providers = new CopyOnWriteArrayList<>();
    private long energy = 0;

    public SimpleBattery(BlockEntityType<?> type, BlockPos pos, BlockState state, long capacity, long inputRate, long outputRate) {
        super(type, pos, state);
        this.capacity = capacity;
        this.inputRate = inputRate;
        this.outputRate = outputRate;
        CoordinateMatchMaker.addProvider(providerDetails);
        CoordinateMatchMaker.addReceiver(receiverDetails);
    }

    @Override
    public long extract(long amount, EnergyReceiver receiver) {
        long extracted = Math.min(Math.min(outputRate, amount), energy);
        energy -= extracted;
        return extracted;
    }

    @Override
    public void exists(EnergyReceiver receiver, ArrayList<Vec3i> pointsOfPresence) {
        receiver.ready(this);
    }

    @Override
    public void ready(EnergyProvider provider) {
        providers.add(provider);
    }

    public static void tick(World world, BlockPos pos, BlockState state, SimpleBattery battery) {
        var leftToFill = Math.min(battery.inputRate, battery.capacity - battery.energy);
        for (EnergyProvider provider : battery.providers) {
            // Less than zero shouldn't be possible, but just in case...
            if (leftToFill <= 0) {
                break;
            }
            long extracted = provider.extract(leftToFill, battery);
            battery.energy += extracted;
            leftToFill -= extracted;
        }
    }

    @Override
    public void unready(EnergyProvider provider) {
        providers.remove(provider);
    }

    @Override
    public void onLoad(BlockEntity blockEntity, ServerWorld serverWorld) {
        CoordinateMatchMaker.addProvider(providerDetails);
        CoordinateMatchMaker.addReceiver(receiverDetails);
    }

    @Override
    public void onUnload(BlockEntity blockEntity, ServerWorld serverWorld) {
        CoordinateMatchMaker.removeProvider(providerDetails);
        CoordinateMatchMaker.removeReceiver(receiverDetails);
        providers.clear();
    }
}
