import arzumify.polyenergy.impl.CoordinateMatchMaker;
import net.minecraft.util.math.Vec3i;
import org.junit.jupiter.api.BeforeEach;
import org.junit.jupiter.api.Test;

import static org.junit.jupiter.api.Assertions.assertEquals;

public class EnergyTests {
    @BeforeEach
    public void setup() {
        // Delete all batteries
        CoordinateMatchMaker.reset();
        System.out.println("Reset CoordinateMatchMaker");
    }

    @Test
    public void testEnergyTransfer() {
        // Create two batteries with coordinates (0, 0, 0) and (1, 0, 0)
        // Battery1 should be able to provide energy to battery2
        var battery1 = new SimpleCodeOnlyBattery(new Vec3i(0, 0, 0), 16, 0, 1, "Battery1");
        var battery2 = new SimpleCodeOnlyBattery(new Vec3i(1, 0, 0), 16, 1, 0, "Battery2");

        // Give battery1 some energy
        battery1.energy = 16;

        // Tick the batteries 10 times
        for (int i = 0; i < 10; i++) {
            SimpleCodeOnlyBattery.tick(battery1);
            SimpleCodeOnlyBattery.tick(battery2);
        }

        // Output the energy of the batteries
        System.out.println("Battery1 energy: " + battery1.energy);
        System.out.println("Battery2 energy: " + battery2.energy);

        // Battery1 should have lost 10 energy and battery2 should have gained 10 energy
        assertEquals(6, battery1.energy);
        assertEquals(10, battery2.energy);
    }

    @Test
    public void testUnloadBattery1() {
        // Create two batteries with coordinates (0, 0, 0) and (1, 0, 0)
        // Battery1 should be able to provide energy to battery2
        var battery1 = new SimpleCodeOnlyBattery(new Vec3i(0, 0, 0), 16, 0, 1, "Battery1");
        var battery2 = new SimpleCodeOnlyBattery(new Vec3i(1, 0, 0), 16, 1, 0, "Battery2");

        // Give battery1 some energy
        battery1.energy = 16;

        // Tick 5 times
        for (int i = 0; i < 5; i++) {
            SimpleCodeOnlyBattery.tick(battery1);
            SimpleCodeOnlyBattery.tick(battery2);
        }

        // Unload battery1
        System.out.println("Unloading battery1");
        battery1.onUnload(null, null);

        // Tick 5 more times
        for (int i = 0; i < 5; i++) {
            SimpleCodeOnlyBattery.tick(battery1);
            SimpleCodeOnlyBattery.tick(battery2);
        }

        // Output the energy of the batteries
        System.out.println("Battery1 energy: " + battery1.energy);
        System.out.println("Battery2 energy: " + battery2.energy);

        // Battery1 should have lost 5 energy and battery2 should have gained 5 energy
        assertEquals(11, battery1.energy);
        assertEquals(5, battery2.energy);
    }

    @Test
    public void testUnloadBattery2() {
        // Create two batteries with coordinates (0, 0, 0) and (1, 0, 0)
        // Battery1 should be able to provide energy to battery2
        var battery1 = new SimpleCodeOnlyBattery(new Vec3i(0, 0, 0), 16, 0, 1, "Battery1");
        var battery2 = new SimpleCodeOnlyBattery(new Vec3i(1, 0, 0), 16, 1, 0, "Battery2");

        // Give battery1 some energy
        battery1.energy = 16;

        // Tick 5 times
        for (int i = 0; i < 5; i++) {
            SimpleCodeOnlyBattery.tick(battery1);
            SimpleCodeOnlyBattery.tick(battery2);
        }

        // Unload battery2
        System.out.println("Unloading battery2");
        battery2.onUnload(null, null);

        // Tick 5 more times
        for (int i = 0; i < 5; i++) {
            SimpleCodeOnlyBattery.tick(battery1);
            SimpleCodeOnlyBattery.tick(battery2);
        }

        // Output the energy of the batteries
        System.out.println("Battery1 energy: " + battery1.energy);
        System.out.println("Battery2 energy: " + battery2.energy);

        // Battery1 should have lost 5 energy and battery2 should have gained 5 energy
        assertEquals(11, battery1.energy);
        assertEquals(5, battery2.energy);
    }

    @Test
    public void testUnloadThenReloadBattery1() {
        // Create two batteries with coordinates (0, 0, 0) and (1, 0, 0)
        // Battery1 should be able to provide energy to battery2
        var battery1 = new SimpleCodeOnlyBattery(new Vec3i(0, 0, 0), 16, 0, 1, "Battery1");
        var battery2 = new SimpleCodeOnlyBattery(new Vec3i(1, 0, 0), 16, 1, 0, "Battery2");

        // Give battery1 some energy
        battery1.energy = 16;

        // Tick 5 times
        for (int i = 0; i < 5; i++) {
            SimpleCodeOnlyBattery.tick(battery1);
            SimpleCodeOnlyBattery.tick(battery2);
        }

        // Unload battery1
        System.out.println("Unloading battery1");
        battery1.onUnload(null, null);

        // Tick 5 more times
        for (int i = 0; i < 5; i++) {
            SimpleCodeOnlyBattery.tick(battery1);
            SimpleCodeOnlyBattery.tick(battery2);
        }

        // Reload battery1
        System.out.println("Reloading battery1");
        battery1.onLoad(null, null);

        // Tick 5 more times
        for (int i = 0; i < 5; i++) {
            SimpleCodeOnlyBattery.tick(battery1);
            SimpleCodeOnlyBattery.tick(battery2);
        }

        // Output the energy of the batteries
        System.out.println("Battery1 energy: " + battery1.energy);
        System.out.println("Battery2 energy: " + battery2.energy);

        // Battery1 should have lost 10 energy and battery2 should have gained 10 energy
        assertEquals(6, battery1.energy);
        assertEquals(10, battery2.energy);
    }

    @Test
    public void testUnloadThenReloadBattery2() {
        // Create two batteries with coordinates (0, 0, 0) and (1, 0, 0)
        // Battery1 should be able to provide energy to battery2
        var battery1 = new SimpleCodeOnlyBattery(new Vec3i(0, 0, 0), 16, 0, 1, "Battery1");
        var battery2 = new SimpleCodeOnlyBattery(new Vec3i(1, 0, 0), 16, 1, 0, "Battery2");

        // Give battery1 some energy
        battery1.energy = 16;

        // Tick 5 times
        for (int i = 0; i < 5; i++) {
            SimpleCodeOnlyBattery.tick(battery1);
            SimpleCodeOnlyBattery.tick(battery2);
        }

        // Unload battery2
        System.out.println("Unloading battery2");
        battery2.onUnload(null, null);

        // Tick 5 more times
        for (int i = 0; i < 5; i++) {
            SimpleCodeOnlyBattery.tick(battery1);
            SimpleCodeOnlyBattery.tick(battery2);
        }

        // Reload battery2
        System.out.println("Reloading battery2");
        battery2.onLoad(null, null);

        // Tick 5 more times
        for (int i = 0; i < 5; i++) {
            SimpleCodeOnlyBattery.tick(battery1);
            SimpleCodeOnlyBattery.tick(battery2);
        }

        // Output the energy of the batteries
        System.out.println("Battery1 energy: " + battery1.energy);
        System.out.println("Battery2 energy: " + battery2.energy);

        // Battery1 should have lost 10 energy and battery2 should have gained 10 energy
        assertEquals(6, battery1.energy);
        assertEquals(10, battery2.energy);
    }

    @Test
    public void twoHalfFullBatteries() {
        System.out.println("Two half-full batteries");
        // Create two batteries with coordinates (0, 0, 0) and (1, 0, 0)
        // They should both provide energy to each other and stay at equilibrium
        var battery1 = new SimpleCodeOnlyBattery(new Vec3i(0, 0, 0), 16, 1, 1, "Battery1");
        var battery2 = new SimpleCodeOnlyBattery(new Vec3i(1, 0, 0), 16, 1, 1, "Battery2");

        // Give battery1 and battery2 some energy
        battery1.energy = 8;
        battery2.energy = 8;

        // Tick 10 times
        for (int i = 0; i < 10; i++) {
            SimpleCodeOnlyBattery.tick(battery1);
            SimpleCodeOnlyBattery.tick(battery2);
        }

        // Output the energy of the batteries
        System.out.println("Battery1 energy: " + battery1.energy);
        System.out.println("Battery2 energy: " + battery2.energy);

        // Battery1 and battery2 should have lost no energy
        assertEquals(8, battery1.energy);
        assertEquals(8, battery2.energy);
    }

    @Test
    public void middleOneCharges() {
        // Create three batteries with coordinates (0, 0, 0), (1, 0, 0), and (2, 0, 0)
        // Battery2 should provide energy to battery1 and battery3
        var battery1 = new SimpleCodeOnlyBattery(new Vec3i(0, 0, 0), 16, 1, 0, "Battery1");
        var battery2 = new SimpleCodeOnlyBattery(new Vec3i(1, 0, 0), 16, 0, 1, "Battery2");
        var battery3 = new SimpleCodeOnlyBattery(new Vec3i(2, 0, 0), 16, 1, 0, "Battery3");

        // Give battery2 some energy
        battery2.energy = 16;

        // Tick 10 times
        for (int i = 0; i < 10; i++) {
            SimpleCodeOnlyBattery.tick(battery1);
            SimpleCodeOnlyBattery.tick(battery2);
            SimpleCodeOnlyBattery.tick(battery3);
        }

        // Output the energy of the batteries
        System.out.println("Battery1 energy: " + battery1.energy);
        System.out.println("Battery2 energy: " + battery2.energy);
        System.out.println("Battery3 energy: " + battery3.energy);

        // Battery1 and battery3 should have gained 8 energy and battery2 should have lost 16 energy
        assertEquals(8, battery1.energy);
        assertEquals(0, battery2.energy);
        assertEquals(8, battery3.energy);
    }

    @Test
    public void middleOneReceives() {
        // Create three batteries with coordinates (0, 0, 0), (1, 0, 0), and (2, 0, 0)
        // Battery1 and battery3 should provide energy to battery2
        var battery1 = new SimpleCodeOnlyBattery(new Vec3i(0, 0, 0), 16, 0, 1, "Battery1");
        var battery2 = new SimpleCodeOnlyBattery(new Vec3i(1, 0, 0), 16, 1, 0, "Battery2");
        var battery3 = new SimpleCodeOnlyBattery(new Vec3i(2, 0, 0), 16, 0, 1, "Battery3");

        // Give battery1 and battery3 some energy
        battery1.energy = 8;
        battery3.energy = 8;

        // Tick 10 times
        for (int i = 0; i < 16; i++) {
            SimpleCodeOnlyBattery.tick(battery1);
            SimpleCodeOnlyBattery.tick(battery2);
            SimpleCodeOnlyBattery.tick(battery3);
        }

        // Output the energy of the batteries
        System.out.println("Battery1 energy: " + battery1.energy);
        System.out.println("Battery2 energy: " + battery2.energy);
        System.out.println("Battery3 energy: " + battery3.energy);

        // Battery2 should have 16 energy and battery1 and battery3 should have 0 energy
        assertEquals(0, battery1.energy);
        assertEquals(16, battery2.energy);
        assertEquals(0, battery3.energy);
    }

    @Test
    public void testEnergyTransferWithRange() {
        // Create two batteries with coordinates (0, 0, 0) and (2, 0, 0)
        // Battery1 should be able to provide energy to battery2
        var battery1 = new SimpleCodeOnlyBattery(new Vec3i(0, 0, 0), 16, 0, 1, "Battery1");
        var battery2 = new SimpleCodeOnlyBattery(new Vec3i(2, 0, 0), 16, 1, 0, "Battery2");

        // Give battery1 some energy
        battery1.energy = 16;

        // Tick the batteries 10 times
        for (int i = 0; i < 10; i++) {
            SimpleCodeOnlyBattery.tick(battery1);
            SimpleCodeOnlyBattery.tick(battery2);
        }

        // Output the energy of the batteries
        System.out.println("Battery1 energy: " + battery1.energy);
        System.out.println("Battery2 energy: " + battery2.energy);

        // Battery1 should have lost no energy and battery2 should have gained no energy
        assertEquals(16, battery1.energy);
        assertEquals(0, battery2.energy);
    }

    @Test
    public void stressTest() {
        // Create 1000 batteries with coordinates (0, 0, 0) to (999, 0, 0)
        // Each battery should be able to provide energy to the next battery
        var batteries = new SimpleCodeOnlyBattery[1000];
        for (int i = 0; i < 1000; i++) {
            batteries[i] = new SimpleCodeOnlyBattery(new Vec3i(i, 0, 0), 16, 1, 2, "Battery" + i);
        }

        // Give the first battery some energy
        batteries[0].energy = 16;

        // Tick once
        for (SimpleCodeOnlyBattery battery : batteries) {
            SimpleCodeOnlyBattery.tick(battery);
        }
    }
}
